/*
=====================================================================

    segkh: Sequence segmentation with recurrent sources
	
    Copyright (C) 2004 Aristides Gionis

    This code is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY.  For more details see the license 
    that accompanies the code.

    The code is an implementation of the algorithms described in 
    the paper

    [1] A. Gionis, H. Mannila, Finding recurrent sources in 
        sequences, 7th International Conference on Research 
        in Computational Molecular Biology (RECOMB) 2003

=====================================================================
*/

#include <stdlib.h>
#include <iostream.h>   // cout, cerr
#include <float.h>      // FLT_MAX
#include <math.h>       // fabs, log

typedef enum {R, C, E, M} _algo;

#define MAX(A,B) ((A)>(B)?(A):(B))
#define MIN(A,B) ((A)<(B)?(A):(B))

struct point {
  float *x;
  point(float *, int);
  point(point *, int);
  point(int);
  ~point();
};

point::point(float *_x, int d)
{
  x = new float[d];
  for (int i=0; i<d; i++)
    x[i] = _x[i];
}

point::point(point *p, int d)
{
  x = new float[d];
  for (int i=0; i<d; i++)
    x[i] = p->x[i];
}

point::point(int d)
{
  x = new float[d];
  for (int i=0; i<d; i++)
    x[i] = 0;
}


point::~point()
{
  delete[] x;
}

void print_point(point* p, int d)
{
  if (p==0)
    return;
  cerr << "(..omitted..)";
  /*
  cerr << "(" << p->x[0];
  for (int j=1; j<d; j++)
    cerr << ", " << p->x[j];
  cerr << ")";
  */
}

void copy_point(point *dest, point *src, int d)
{
  for (int j=0; j<d; j++)
    dest->x[j] = src->x[j];
}


void add_point(point *dst, point *src, int d)
{
  for (int j=0; j<d; j++)
    dst->x[j] += src->x[j];
}


void multiply_point(point *dst, float x, int d)
{
  for (int j=0; j<d; j++)
    dst->x[j] *= x;
}


void zero_point(point *p, int d)
{
  for (int j=0; j<d; j++)
    p->x[j] = 0.0;
}

float distance_sq(point *p, point *q, int d)
{
  float dist_sq = 0.0;
  for (int j=0; j<d; j++)
    dist_sq += (p->x[j] - q->x[j]) * (p->x[j] - q->x[j]);
  return dist_sq;
}



float minus_log_likehood(point** A, int n, int d, int k, int h, 
			 int* seg_start, point** level_val, int *level_id)
{  
  float error = 0.0;
  for(int l=0; l<d; l++) {
    float acc_sum[n], acc_sum_sq[n];

    acc_sum[0] =  A[0]->x[l];
    acc_sum_sq[0] = (A[0]->x[l])*(A[0]->x[l]);
    
    for (int i=1; i<n; i++) {
      acc_sum[i] = acc_sum[i-1] + A[i]->x[l];
      acc_sum_sq[i] = acc_sum_sq[i-1] + (A[i]->x[l])*(A[i]->x[l]);
    }
      
    for(int i=0; i<k; i++) {
      int start = seg_start[i];
      int end = (i==k-1) ? (n-1) :  seg_start[i+1]-1;
      int level = level_id[i];
      point *val = level_val[level];
      
      float sum, sum_sq;
      if (i>0) {
	sum    = acc_sum[end]    - acc_sum[start-1];
	sum_sq = acc_sum_sq[end] - acc_sum_sq[start-1];
      } else {
	sum    = acc_sum[end];
	sum_sq = acc_sum_sq[end];
      }
      int num = end-start+1;

      error += sum_sq + val->x[l] * ( num * val->x[l] - 2 * sum );
    }
  }
  return error;
}


float local_search_clustering(point** A, float *w, int n, int d, int k, point** C)
  /*
    Local search algorithm for clustering n points of dimension d
    int k clusters

    The objective is weighted sum of squares of distances of each point, 
    to its closest center

    The algorithm is based on a local search heuristic and it is analyzed by
    Kanungo et al, at SoCG 2002. 

    Approximation guarantee is 25 :(

    The algorithm assumes that memory is already initialized for (point** C) 
  */
{
  // is_center[i] is -1 if point i is selected as center, 0 o/w
  // is_center[i] is shows which center is point i
  //
  // assigned_to[i] shows which center point i is assigned to 
  //
  // center[i] shows which point is center i (the inverse of is_center)

  int is_center[n], assigned_to[n], center[k];

  // Initialize centers arbitrary with the first k points
  for (int i=0; i<MIN(n,k); i++) {
    copy_point(C[i], A[i], d);
    is_center[i] = i;
    assigned_to[i] = i;
    center[i] = i;
  }

  if (n<=k)
    // nothing to cluster !
    return 0.0;

  // Compute which center each point is assign to, for the rest of the points
  //
  // Cost keeps the cost of the current solution
  // this is weighted sum of squares of distance of each point
  // to its closest center. 

  float Cost = 0.0;
  for (int i=k; i<n; i++) {
    is_center[i] = -1;
    float min_dist = FLT_MAX;
    int closest_center = -1;
    for (int j=0; j<k; j++) {
      float dist = distance_sq(A[i], A[center[j]], d);
      if (dist < min_dist) {
	min_dist = dist;
	closest_center = center[j];
      }
    }
    Cost += w[i] * min_dist;
    assigned_to[i] = closest_center; 
  }

  // Main local search loop of the algorithm

  int stable = 0;
  while (stable == 0) {
    stable = 1;
    for (int i=0; i<k; i++) 
      for (int j=0; j<n; j++) 
	if (is_center[j] == -1) {
	  // consider replacing center[i] with point A[j]
	  float cost_difference = 0.0;
	  int new_assigned_to[n];

	  // Compute the cost difference of the swap
	  for (int r=0; r<n; r++) {

	    float dist_rc = distance_sq(A[r], A[assigned_to[r]], d);
	    if (assigned_to[r] == center[i]) {
	      // Point r is assigned to center[i] by the current solution
	      // Find which center it should be assigned in the new solution
	      // Condidates are the rest k-1 centers and candidate center point j
	      float min_dist = distance_sq(A[r], A[j], d);
	      int closest_center = j;
	      for (int c=0; c<k; c++) {
		if (c != i) {
		  float dist = distance_sq(A[r], A[center[c]], d);
		  if (dist < min_dist) {
		    min_dist = dist;
		    closest_center = center[c];
		  }
		}
	      }
	      cost_difference += w[r] * (min_dist - dist_rc);
	      new_assigned_to[r] = closest_center;
	    } else {
	      // Point r is assigned to some center other than center[i] 
	      // Check if point r should be reassigned to candidate center point j
	      float dist_rj = distance_sq(A[r], A[j], d);
	      if (dist_rj < dist_rc) {
		cost_difference += w[r] * (dist_rj - dist_rc);
		new_assigned_to[r] = j;
	      } else 
		new_assigned_to[r] = assigned_to[r];
	    }
	  }

	  // If the difference is negative make the swap
	  if (cost_difference < 0) {
	    is_center[center[i]] = -1;
	    is_center[j] = i;
	    center[i] = j;
	    for (int r=0; r<n; r++) 
	      assigned_to[r] = new_assigned_to[r];
	    Cost += cost_difference;
	    stable = 0;
	    // do not look for other swaps for center[i] in that loop  
	    break;
	  }

	}    
  }

  // Compute centroids
  int cluster_sz[k];
  for (int  i=0; i<k; i++) {
    cluster_sz[i] = 0;
    zero_point(C[i], d);
  }
  
  for (int i=0; i<n; i++) {
    add_point(C[is_center[assigned_to[i]]], A[i], d);
    cluster_sz[is_center[assigned_to[i]]]++;
  }

  for (int i=0; i<k; i++)
    multiply_point(C[i], 1/((float)cluster_sz[i]), d);
  
  // Compute and return cost
  Cost = 0.0;
  for (int i=0; i<n; i++) 
    Cost += distance_sq(A[i], C[is_center[assigned_to[i]]], d);
  return Cost;
}



float sum_sq_error(point** acc_sum, point** acc_sum_sq, float *acc_w, int n, int d, int i, int j)
// weighted sum of squares error from i to j entries of the table, INCLUDING both i and j
{
  if (j<i || j<0 || j>=n) {
    cerr << "error in sum_sq_error indexing" << endl;
    exit(0);
  }
  float sum[d], sum_sq[d], sum_w;
  if (i>0) {
    for (int l=0; l<d; l++) {
      sum[l]    = acc_sum[j]->x[l]    - acc_sum[i-1]->x[l];
      sum_sq[l] = acc_sum_sq[j]->x[l] - acc_sum_sq[i-1]->x[l];
    }
    sum_w = acc_w[j] - acc_w[i-1];
  } else {
    for (int l=0; l<d; l++) {
      sum[l]    = acc_sum[j]->x[l];
      sum_sq[l] = acc_sum_sq[j]->x[l];
    }
    sum_w  = acc_w[j];
  }  

  float val = 0.0;
  for (int l=0; l<d; l++) 
    val += sum_sq[l] - sum[l]*sum[l]/sum_w;
  
  return val;
}


void avg(point** acc_sum, float *acc_w, int n, int d, int i, int j, point* &result)
// weighted average from i to j entries of the table, INCLUDING both i and j
{
  if (j<i || j<0 || j>=n) {
    cerr << "error in avg indexing" << endl;
    exit(0);
  }
  float sum_w;
  if (i>0) {
    sum_w = acc_w[j] - acc_w[i-1];
    for (int l=0; l<d; l++) 
      result->x[l] = (acc_sum[j]->x[l] - acc_sum[i-1]->x[l]) / sum_w;
  } else {
    sum_w = acc_w[j];
    for (int l=0; l<d; l++) 
      result->x[l] = acc_sum[j]->x[l] / sum_w;
  }
}


void avg(point** acc_sum, int n, int d, int i, int j, point* &result)
// average from i to j entries of the table, INCLUDING both i and j
{
 if (j<i || j<0 || j>=n) {
    cerr << "error in avg indexing" << endl;
    exit(0);
  }

 int tot = j-i+1;
 if (i>0)
   for (int l=0; l<d; l++) 
     result->x[l] = (acc_sum[j]->x[l] - acc_sum[i-1]->x[l]) / tot;
 else 
   for (int l=0; l<d; l++) 
     result->x[l] = acc_sum[j]->x[l] / tot;
}


float dyn_k_k(point **A, int n, int d, int k, int *seg_start, point **seg_val, const float *_w = 0)
{
  // Remember that indexing goes from 0 to k-1 and from 0 to n-1 !!

  //  float Table_val[k][n];
  //  int   Table_ptr[k][n];

  float **Table_val = new float*[k];
  for (int i=0; i<k; i++)
    Table_val[i] = new float[n];

  int **Table_ptr = new int*[k];
  for (int i=0; i<k; i++)
    Table_ptr[i] = new int[n];

  // Compute cummulative sums and sums of squares
  // They'd be useful for computing Sum(xi - xm) = sum(xi) - (sum(xi))/n

  point *acc_sum[n], *acc_sum_sq[n];
  for (int i=0; i<n; i++) {
    acc_sum[i] = new point(d);
    acc_sum_sq[i] = new point(d);
  } 
  float acc_w[n], w[n];
  
  if (_w==0) 
    for(int j=0; j<n; j++)  w[j] = 1.0;
  else 
    for(int j=0; j<n; j++)  w[j] = _w[j];

  for(int i=0; i<d; i++) {
    acc_sum[0]->x[i] = w[0]*(A[0]->x[i]);
    acc_sum_sq[0]->x[i] = w[0]*(A[0]->x[i])*(A[0]->x[i]);
  }
  acc_w[0] = w[0];
    
  for (int j=1; j<n; j++) {
    for(int i=0; i<d; i++) {
      acc_sum[j]->x[i] = (acc_sum[j-1]->x[i]) + w[j]*(A[j]->x[i]);
      acc_sum_sq[j]->x[i] = (acc_sum_sq[j-1]->x[i]) + w[j]*(A[j]->x[i])*(A[j]->x[i]);
    }
    acc_w[j] = acc_w[j-1] + w[j];
  }

  // Initialize the first row

  for (int j=0; j<n; j++) {
    Table_val[0][j] = sum_sq_error(acc_sum, acc_sum_sq, acc_w, n, d, 0, j);
    Table_ptr[0][j] = -1;
  }

  // Start the Dynamic Programming !

  for (int i=1; i<k; i++) 
    for (int j=0; j<n; j++) {
      // This corresponds to having just one segment
      float best_val = Table_val[0][j];
      int best_ptr = -1;
      // Search for the best solution with i segments, using i-1 segments plus 1
      for (int x=1; x<=j; x++) {
	float val = Table_val[i-1][x-1] + sum_sq_error(acc_sum, acc_sum_sq, acc_w, n, d, x, j);
	if (val <= best_val) {
	  best_val = val;
	  best_ptr = x-1;
	}
      }
      Table_val[i][j] = best_val;
      Table_ptr[i][j] = best_ptr;
    }

  // Trace the solution backwards 

  int ptr = n-1;
  for (int i=k-1; i>=0; i--) {
    int new_ptr = Table_ptr[i][ptr];
    seg_start[i] = new_ptr + 1;
    avg(acc_sum, acc_w, n, d, new_ptr + 1, ptr, seg_val[i]);
    ptr = MAX(new_ptr,0);
  }

  for (int i=0; i<n; i++) {
    delete acc_sum[i];
    delete acc_sum_sq[i];
  } 

  float ret_val = Table_val[k-1][n-1];

  for (int i=0; i<k; i++)
    delete[] Table_val[i];
  delete[] Table_val;

  for (int i=0; i<k; i++)
    delete[] Table_ptr[i];
  delete[] Table_ptr;
  
  return ret_val;
}


float dyn_n_h(point** A, int n, int d, int h, point** level_val, const float* _w = 0)
  /*
    In high dimension this is just clustering !
  */
{
  float w[n];  
  if (_w==0) 
    for(int j=0; j<n; j++)  w[j] = 1.0;
  else 
    for(int j=0; j<n; j++)  w[j] = _w[j];

  return local_search_clustering(A, w, n, d, h, level_val);
  // The above returns discrete solution
  // Need also to readjust the solution at centroids of clusters
}


float assign_to_levels_and_update(point** A, int n, int d, int k, int h, 
				  int* seg_start, point** seg_val, point** level_val,
				  int *level_id)
  /*
    Given segments described by  (int *seg_start, point** seg_val)
    and given levels described by (point **level_val)
    
    it rounds each segment to its closest level

    After that it moves the levels so that they are the new averages
    of the points assigned to them.

    It also computes (int* level_id), which has size k and indicates which 
    level is assigned to each segment

    Finally the function computes and returns the total error after the rounding process
  */
{
  int level_sz[h];
  point* new_level_val[h];

  // Initialize

  for(int j=0; j<h; j++) {
    level_sz[j] = 0;
    new_level_val[j] = new point(d);
  }
  
  for(int i=0; i<k; i++) {
    // find the nearest level
    int nearest_level = -1;
    float nearest_level_dist = FLT_MAX;
    for(int  j=0; j<h; j++) {
      float dist = distance_sq(seg_val[i], level_val[j], d);
      if (dist < nearest_level_dist) {
	nearest_level_dist = dist;
	nearest_level = j;
      }
    }
    // update the level info
    level_id[i] = nearest_level;
    int seg_size = (i<k-1) ? seg_start[i+1]-seg_start[i] : n-seg_start[i];
    level_sz[nearest_level] += seg_size;
    for(int c=0; c<d; c++)
      new_level_val[nearest_level]->x[c] += seg_size * seg_val[i]->x[c];
  }

  // compute new levels

  for(int j=0; j<h; j++) {
    if (level_sz[j] > 0.0)
      for(int c=0; c<d; c++)
	level_val[j]->x[c] = new_level_val[j]->x[c] /(float) level_sz[j];
    else 
      for(int c=0; c<d; c++)
	level_val[j]->x[c] = FLT_MAX;
  }
  
  // Compute error

  float error = 0.0;
  for(int i=0; i<k; i++) {
    point* level = level_val[level_id[i]];
    int start = seg_start[i];
    int end = (i<k-1) ? seg_start[i+1] : n;
    for (int j=start; j<end; j++)
      error += distance_sq(A[j], level, d);
  }

  // Deallocate memory 

  for(int j=0; j<h; j++) 
    delete new_level_val[j];
 
  return error;
}

float rounding_algo(point** A, int n, int d, int k, int h)
{
  cerr << endl << "Rounding Algorithm" << endl;
  cerr         << "------------------" << endl;

  int seg_start[k];

  point* seg_val[k];
  for (int i=0; i<k; i++) 
    seg_val[i] = new point(d);

  point* level_val[h];
  for (int i=0; i<h; i++) 
    level_val[i] = new point(d);

  // Find optimal segments solving dynamic programming

  float kkerror = dyn_k_k(A, n, d, k, seg_start, seg_val);

  // Find optimal levels by doing clustering

  float nherror = dyn_n_h(A, n, d, h, level_val);
 
  cerr << endl << "(k,k) Segments: " << endl;
  for(int i=0; i<k; i++) {
    cerr << i << ".  start: " << seg_start[i] << "  value: ";
    print_point(seg_val[i], d);
    cerr << endl;
  }
  cerr << "(k,k) error: " << kkerror << endl;
 
  cerr << endl << "(n,h) Levels (Clustering): " << endl;
  for(int j=0; j<h; j++) {
    cerr << "level " << j << ": ";
    print_point(level_val[j], d);
    cerr << endl;
  }
  cerr << "(n,h) error: " << nherror << endl;

  // Find the nearest level for each segment
  // Update the levels at the same time

  int level_id[k];

  float kherror = assign_to_levels_and_update(A, n, d, k, h, seg_start, seg_val, level_val, level_id);
 
  // output segmentation description

  cerr << endl << "(k,h) segmentation: " << endl ;
  for(int i=0; i<k; i++) {
    int level = level_id[i];
    cerr << i << ".  start: " << seg_start[i] << "  level: " << level << "  value: ";
    print_point(level_val[level], d);
    cerr << endl;
  }
  cerr << "(k,h) error: " << kherror << endl;  

  float ret_val = 2*minus_log_likehood(A, n, d, k, h, seg_start, level_val, level_id) + (k+h*d)*log(n);

  for (int i=0; i<k; i++) 
    delete seg_val[i];

  for (int i=0; i<h; i++) 
    delete level_val[i];

  return ret_val;
}



float sum_sq_error_to_levels(point** acc_sum, point** acc_sum_sq, int n, int d, int h, point** level_val, int i, int j, int &level)
  /*
    sum of squares error for segment from i to j, INCLUDING both i and j

    The error is measured to the closest level specified at (float *level_val)

    The closest level is returned at variable (int level)
  */
{
  if (j<i || j<0 || j>=n) {
    cerr << "error in sum_sq_error_to_levels indexing" << endl;
    exit(0);
  }

  float sum[d], sum_sq[d];
  if (i>0) 
    for (int l=0; l<d; l++) {
      sum[l]    = acc_sum[j]->x[l]    - acc_sum[i-1]->x[l];
      sum_sq[l] = acc_sum_sq[j]->x[l] - acc_sum_sq[i-1]->x[l];
    }
  else
    for (int l=0; l<d; l++) {
      sum[l]    = acc_sum[j]->x[l];
      sum_sq[l] = acc_sum_sq[j]->x[l];
    }

  float num = (float)j-i+1;

  float minerror = FLT_MAX;
  level = -1;

  for (int l=0; l<h; l++) {
    point* val = level_val[l];
    float error = 0.0;
    for (int j=0; j<d; j++) 
      error += sum_sq[j] + val->x[j] * ( num * val->x[j] - 2 * sum[j] );
    if (error<minerror) {
      minerror = error;
      level = l;
    }
  }
  
  return minerror;
}


float dyn_k_given_h(point** A, int n, int d, int k, int h, point** level_val, int *seg_start, point** seg_avg, int *level_id)
  /*
    Very similar stuff to dyn_k_k

    The main difference how the error is computed from a segment to a given level, 
    that is we are not allowed to use the mean for that segment anymore.

    So instead of using  
    
    Sum(xi - xm) = sum(xi) - (sum(xi))/n

    we have to use

    Sum(xi - l) = sum(xi) + n l - 2 l sum(xi)

    As before:
    (int *seg_start) has the start of each segment,
    (point *seg_avg) has  the average within each segment 
    (notice that in general this is different that the value of level to which the segment is assigned)
    (float *level_id) has the level id to which each segment is assigned 
  */
{

  float **Table_val = new float*[k];
  for (int i=0; i<k; i++)
    Table_val[i] = new float[n];

  int **Table_ptr = new int*[k];
  for (int i=0; i<k; i++)
    Table_ptr[i] = new int[n];

  int **Table_level = new int*[k];
  for (int i=0; i<k; i++)
    Table_level[i] = new int[n];


  // Compute cummulative sums and sums of squares
  // They are useful for computing sum of squares error to the levels

  point *acc_sum[n], *acc_sum_sq[n];
  for (int i=0; i<n; i++) {
    acc_sum[i] = new point(d);
    acc_sum_sq[i] = new point(d);
  } 

  for(int i=0; i<d; i++) {
    acc_sum[0]->x[i] = A[0]->x[i];
    acc_sum_sq[0]->x[i] = (A[0]->x[i])*(A[0]->x[i]);
  }
    
  for (int j=1; j<n; j++) 
    for(int i=0; i<d; i++) {
      acc_sum[j]->x[i] = acc_sum[j-1]->x[i] + A[j]->x[i];
      acc_sum_sq[j]->x[i] = acc_sum_sq[j-1]->x[i] + (A[j]->x[i])*(A[j]->x[i]);
    }

  // Initialize the first row
  
  for (int j=0; j<n; j++) {
    int level;
    Table_val[0][j] = sum_sq_error_to_levels(acc_sum, acc_sum_sq, n, d, h, level_val, 0, j, level);
    Table_ptr[0][j] = -1;
    Table_level[0][j] = level;
  }

  // Start the Dynamic Programming !

  for (int i=1; i<k; i++) 
    for (int j=0; j<n; j++) {
      // This corresponds to having just one segment
      float best_val = Table_val[0][j];
      int best_ptr = -1;
      int best_level =  Table_level[0][j];
      // Search for the best solution with i segments, using i-1 segments plus 1
      for (int x=1; x<=j; x++) {
	int level;
	float val = Table_val[i-1][x-1] + sum_sq_error_to_levels(acc_sum, acc_sum_sq, n, d, h, level_val, x, j, level);
	if (val <= best_val) {
	  best_val = val;
	  best_ptr = x-1;
	  best_level = level;
	}
      }
      Table_val[i][j] = best_val;
      Table_ptr[i][j] = best_ptr;
      Table_level[i][j] = best_ptr;
    }

  // Trace the solution backwards 

  int ptr = n-1;
  for (int i=k-1; i>=0; i--) {
    int new_ptr = Table_ptr[i][ptr];
    seg_start[i] = new_ptr + 1;
    avg(acc_sum, n, d, new_ptr + 1, ptr, seg_avg[i]);
    level_id[i] =  Table_level[i][ptr];
    ptr = MAX(new_ptr,0);
  }

  for (int i=0; i<n; i++) {
    delete acc_sum[i];
    delete acc_sum_sq[i];
  } 

  float ret_val =  Table_val[k-1][n-1];

  for (int i=0; i<k; i++)
    delete []Table_val[i];
  delete[] Table_val;

  for (int i=0; i<k; i++)
    delete[] Table_ptr[i];
  delete[] Table_ptr;

  for (int i=0; i<k; i++)
    delete[] Table_level[i];
  delete[] Table_level;

  return ret_val;
}


float EM_algo(point** A, int n, int d, int k, int h)
{

  cerr << endl << "EM Algorithm" << endl;
  cerr         << "------------" << endl;

  int seg_start[k];   
  int level_id[k];

  point* seg_avg[k];
  for (int i=0; i<k; i++) 
    seg_avg[i] = new point(d);
  
  point* level_val[h];
  for (int i=0; i<h; i++) 
    level_val[i] = new point(d);

  // Get optimal levels

  float nherror = dyn_n_h(A, n, d, h, level_val);
  
  cerr << endl << "Initial levels: " << endl;
  for(int j=0; j<h; j++) {
    cerr << "level " << j << ": ";
    print_point(level_val[j], d);
    cerr << endl;
  }
  cerr << "(n,h) error: " << nherror << endl;

  int iter = 0;
  float kherror, new_kherror = FLT_MAX;

  do {
    kherror = new_kherror;

    cerr << endl << "Iteration " << iter++ << endl;

    // solve for optimal segments GIVEN levels

    dyn_k_given_h(A, n, d, k, h, level_val, seg_start, seg_avg, level_id);

    // Find the nearest level for each segment
    // Update the levels at the same time
    
    new_kherror = assign_to_levels_and_update(A, n, d, k, h, seg_start, seg_avg, level_val, level_id);
    
    // output segmentation description
    
    cerr << endl << "(k,h) segmentation: " << endl ;
    for(int i=0; i<k; i++) {
      int level = level_id[i];
      cerr << i << ".  start: " << seg_start[i] << "  level: " << level << "  value: ";
      print_point(level_val[level], d);
      cerr << endl;
    }
    cerr << "(k,h) error: " << new_kherror << endl;

  } while (new_kherror < kherror);
  
  float ret_val = 2*minus_log_likehood(A, n, d, k, h, seg_start, level_val, level_id) + (k+h*d)*log(n);

  for (int i=0; i<k; i++)
    delete seg_avg[i];
  
  for (int i=0; i<h; i++) 
    delete level_val[i];

  return ret_val;
}


float clustering_algo(point** A, int n, int d, int k, int h)
{
  cerr << endl << "Clustering Algorithm" << endl;
  cerr         << "--------------------" << endl;

  float seg_w[k];
  int seg_start[k];

  point* seg_val[k];
  for (int i=0; i<k; i++) 
    seg_val[i] = new point(d);

  point* level_val[h];
  for (int i=0; i<h; i++) 
    level_val[i] = new point(d);

  // Solve optimal (k,k) to get the segments

  /*  float kkerror = */ dyn_k_k(A, n, d, k, seg_start, seg_val);

  // Compute weights, which are in fact the length of the segments

  for(int i=0; i<k; i++) {
    int end = (i<k-1) ? seg_start[i+1] : n;
    seg_w[i] = end - seg_start[i];
  }

  // Solve optimal (k,h) with k weighted points to get the levels

  /*  float nhwerror = */ dyn_n_h(seg_val, k, d, h, level_val, seg_w);
  /* 
    cerr << endl << "(k,k) Segments: " << endl;
    for(int i=0; i<k; i++) {
    cerr << i << ".  start: " << seg_start[i] << "  value: ";
    print_point(seg_val[i], d);
    cerr << endl;
    }
    cerr << "(k,k) error: " << kkerror << endl;
    
    cerr << endl << "(k,h,w) Levels (Clustering): " << endl;
    for(int j=0; j<h; j++) {
    cerr << "level " << j << ": ";
    print_point(level_val[j], d);
    cerr << endl;
    }
    cerr << "(k,h,w) error: " << nhwerror << endl;
  */
  // Find the nearest level for each segment
  // Update the levels at the same time
  
  int level_id[k];
  
  float kherror = assign_to_levels_and_update(A, n, d, k, h, seg_start, seg_val, level_val, level_id);

  // output segmentation description

  cerr << endl << "(k,h) segmentation: " << endl ;
  for(int i=0; i<k; i++) {
    int level = level_id[i];
    cerr << seg_start[i] << " " << level << endl;
    //    print_point(level_val[level], d);
    //    cerr << endl;
  }
  cerr << "(k,h) error: " << kherror << endl;

  float ret_val = 2*minus_log_likehood(A, n, d, k, h, seg_start, level_val, level_id) + (k+h*d)*log(n);

  for (int i=0; i<k; i++) 
    delete seg_val[i];

  for (int i=0; i<h; i++) 
    delete level_val[i];

  return ret_val;
}


void deallocate(point **A, int n)
{
  for (int i=0; i<n; i++)
    delete A[i];
  delete[] A;
}


void get_input(point** &A, int &n, int &d)
{
  cin >> d;
  n = 0;
  int SZ = 1000;

  A = new point*[SZ];

  float val;
  cin >> val;
  while (!cin.eof()) {
    // if the first coordinate is non-negative, read the rest of the dimensions
    float x[d];
    x[0] = val;
    for (int i=1; i<d; i++) 
      cin >> x[i];

    if (n == SZ) {
      // Array is filled. Allocate another array of double the size
      point **B = new point*[2*SZ];
      for (int i=0; i<n; i++)
	B[i] = new point(A[i], d);
     
      // Deallocate the old array, copy the new to the old, and update size
      deallocate(A, n);
      A = B;
      SZ *= 2;
    }

    // copy the point just read 
    A[n++] = new point(x, d);
    cin >> val;
  }
}



void normalize(point** &A, int n, int d)
{    
  for(int l=0; l<d; l++) {
    float sum = 0.0;
    float sum_sq = 0.0;
    for (int i=0; i<n; i++) {
      sum += A[i]->x[l];
      sum_sq += (A[i]->x[l])*(A[i]->x[l]);
    }
    float stddev = sqrt(sum_sq/(float)n - (sum*sum) / ((float)n*n));
    if (sum_sq!=0.0) 
      for (int i=0; i<n; i++)
	A[i]->x[l] =  A[i]->x[l] / stddev;
  }
}


float segment(point **A, int n, int d, int k, int h, _algo algo)
{
  /*
    float minBIC = FLT_MAX;
    int kbest = 0, hbest = 0;
    for (int ki = 2; ki<=k; ki++) {
    for (int hi = 2; hi<=MIN(h,ki); hi++) {
  */
  float BIC;
  switch (algo){
  case R:
    BIC = rounding_algo(A, n, d, k, h);
    break;
  case C:
    BIC = clustering_algo(A, n, d, k, h);
    break;
  case E:
    BIC = EM_algo(A, n, d, k, h);
    break;
      default:
	break;
  }
  cerr << endl << "k, h: " << k << ", " << h <<  "  BIC: " << BIC << endl;
  /*
    cout << BIC << " ";  
    if(BIC<minBIC) {
    kbest = ki;
    hbest = hi;
    minBIC = BIC;
    }
    }
    cout << endl;
    }
    
    cerr << endl << "Best model (k, h, BIC): (" << kbest << ", " << hbest <<  ", " << minBIC << ")" << endl;
    return minBIC;
  */
  return BIC;
}


int main(int argc, char *argv[])
{
  //
  // The program reads a timeseries from stdin
  // First it reads d (number of dimensions)
  // and, then the points, 
  // one line per point, d numbers per line,
  //
  // k and h are read from the command line
  //
  // NOTICE: It assume correct input files!! 
  // It doesn't check for rows with less that d numbers etc.
  //

  if (argc != 4) {
    cerr << endl 
         << "Usage: segkh r/c/e k h" << endl << endl
         << "  r: Rounding algorithm" << endl
         << "  c: Clustering algorithm" << endl
         << "  e: EM algorithm" << endl << endl
      //         << "  m: EM algorithm (start with levels found in clustering algorithm)" << endl << endl
         << "  k: number of segments" << endl
         << "  h: number of levels" << endl << endl
         << "  Input from stdin" << endl << endl
         << "Example: segkh c 10 4 < timeseries.1" << endl << endl;
    return 0;
  }

  _algo algo;
  switch (argv[1][0]){
  case 'r':
    algo = R;
    break;
  case 'c':
    algo = C;
    break;
  case 'e':
    algo = E;
    break;
    //  case 'm':
    //    algo = M;
    //    break;
  default:
    break;
  }

  int k = atoi(argv[2]), h = atoi(argv[3]);

  point **A;
  int n, d;
  get_input(A, n, d);
  //  normalize(A, n, d);

  segment(A, n, d, k, h, algo);

  deallocate(A, n);

  cerr << endl;
  return 0;
}

